(:~
 : This modules handles everything concerning the critical view of the BdN
 : website. Its main function, critical:transform, is triggered by
 : xml-utils:make-html-critical-view.
 :
 : @author Hannes Riebl
 : @author Michelle Weidling
 :)

xquery version "3.1";

module namespace critical="http://bdn-edition.de/ns/critical";

declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace xhtml="http://www.w3.org/1999/xhtml";

import module namespace commons="http://bdn-edition.de/ns/commons" at "commons.xqm";


(:~
 : The main transformation routine. It iterates all nodes of a given TEI/XML
 : file in the BdN specific intermediate format and procudes the HTML output for
 : the critical view.
 :
 : This function is called by xml-utils:make-html-critical-view.
 :
 : @author Hannes Riebl
 : @author Michelle Weidling
 : @version 1.0
 : @param $nodes the nodes of a given TEI/XML document
 : @param $filename e.g. "2228v-if.xml"
 : @return the respective HTML nodes
 :)
declare function critical:transform
(
  $nodes as node()*,
  $display-markers as xs:boolean,
  $rdg-map as map(*)
) as node()*
{
  for $node in $nodes
  return
    typeswitch($node)
    case text() return
      text{replace($node, "@", " ")}

    case comment() return
      ()

    case element(tei:teiHeader) return
      ()

    case element(tei:note) return
      if ($node/@type="editorial-commentary") then
        ()
      else
        (element xhtml:div
        {
          attribute class {"note"},
          critical:transform($node/node(), true(), $rdg-map)
        },
        (: Griesbach has notes that don't belong to a specific chapter.
        these need a critical apparatus too :)
        if (not($node/ancestor::tei:div[@type = "section"])) then
          critical:make-apparatus($node, $rdg-map)
        else
          ())

    case element(tei:front) return
      element xhtml:div
      {
        attribute class {"front"},
        critical:transform($node/node(), true(), $rdg-map)
      }

    case element(tei:titlePage) return
      element xhtml:span
      {
        attribute class {"title-page"},
        critical:transform($node/node(), true(), $rdg-map)
      }

    case element(tei:titlePart) return
      if ($node/@type = "main") then
        if ($node/ancestor::tei:rdg[@type = ("ppl", "ptl")]) then
          element xhtml:span
          {
            attribute class {"rdg-head"},
            critical:transform($node/node(), true(), $rdg-map)
          }
        else
          element xhtml:h1
          {
            attribute class {"main-title"},
            critical:transform($node/node(), true(), $rdg-map)
          }
      else if ($node/@type = "sub") then
        if ($node/ancestor::tei:rdg[@type = ("v", "pt", "pp")]) then
          element xhtml:span
          {
            attribute class {"rdg-head"},
            critical:transform($node/node(), true(), $rdg-map)
          }
        else
          element xhtml:h2
          {
            attribute class {"sub-title"},
            critical:transform($node/node(), true(), $rdg-map)
          }
      else if ($node/@type = "volume") then
        element xhtml:span
        {
          attribute class {"volume"},
          critical:transform($node/node(), true(), $rdg-map)
        }
      else
        ()

    case element(tei:div) return
      if ($node[@type = ("section", "contents", "preface", "corrigenda", "split", "entry")]) then
      ( if ($node/@xml:id) then
          commons:make-text-anchor(string($node/@xml:id))
        else
          (),
        element xhtml:div
        {
          attribute class {$node/@type},
          critical:transform($node/node(), true(), $rdg-map)
        },
        if (not($node/ancestor::tei:rdg)) then
          critical:make-apparatus($node, $rdg-map)
        else
          ()
      )
      else if ($node[@type = "synopsis"]) then
        (if ($node/@xml:id) then
          commons:make-text-anchor(string($node/@xml:id))
        else
          (),
        critical:transform($node/tei:head, true(), $rdg-map),
        <table>
          <tr>
            <td>
              {critical:transform($node/tei:div[@type = 'letter'][1]/tei:head, true(), $rdg-map)}
            </td>
            <td>
              {critical:transform($node/tei:div[@type = 'letter'][2]/tei:head, true(), $rdg-map)}
            </td>
          </tr>
          {   (: ATTENTION: this implies that the two div[@type = 'letter'] always have the same number of ab :)
            for $ab in $node/descendant::tei:div[@type = 'letter'][1]/tei:ab
              let $corresp :=
                $node/tei:div[@type = 'letter'][2]/tei:ab[@xml:id = $ab/@corresp]
              return
                <tr>
                  <td class="syn-td">
                    {critical:transform($ab, true(), $rdg-map)}
                  </td>
                  <td class="syn-td">
                    {critical:transform($corresp, true(), $rdg-map)}
                  </td>
                </tr>
          }
        </table>)

      else
        (if ($node/@xml:id) then
          commons:make-text-anchor(string($node/@xml:id))
        else
          (),
        critical:transform($node/node(), true(), $rdg-map))

    case element(tei:ab) return
      if ($node[@type = "half-title"]) then
        element xhtml:div
        {
          attribute class {$node/@type},
          critical:transform($node/node(), true(), $rdg-map)
        }
      else
        (critical:transform($node/node(), true(), $rdg-map),
        commons:make-whitespace-after($node))

    case element(tei:p) return
      if ($node/@rend = "center-aligned") then
        element xhtml:div
        {
          attribute class {$node/@rend},
          critical:transform($node/node(), true(), $rdg-map)
        }
      else
        element xhtml:p
        {
          if ($node/@rend) then
            attribute class {$node/@rend}
          else
            if ($node/parent::tei:rdg
            and $node/ancestor::tei:app/preceding-sibling::*[1][self::tei:p]) then
              attribute class {"indent"}
            else if ($node/preceding-sibling::tei:p
              or $node/preceding-sibling::*[1][self::tei:app]/tei:lem/child::*[last()][self::tei:list]
              or $node/parent::tei:lem and $node/ancestor::tei:app[1]/preceding-sibling::*[1][self::tei:p]
              or $node/preceding::tei:p
                and $node/preceding::tei:p/ancestor::tei:div[@type = 'section']
                = $node/ancestor::tei:div[@type = 'section'] and not($node/parent::tei:note)) then
              attribute class {"indent"}
            else
              attribute class {"no-indent"},
          critical:transform($node/node(), true(), $rdg-map)
        }

    case element(tei:head) return
      switch($node/@type)
      case "main" return
        element xhtml:div
        {
          attribute class {"list-main-head"},
          critical:transform($node/node(), true(), $rdg-map)
        }
      case "sub" return
        element xhtml:div
        {
          attribute class {"list-sub-head"},
          critical:transform($node/node(), true(), $rdg-map)
        }
      default return
        (
          element xhtml:h2
          {
            critical:transform($node/node(), true(), $rdg-map)
          },
          if ($node[descendant::tei:rdg[@type = ("pp", "pt", "v")] and not(ancestor::tei:div[@type = "section"])]) then
            critical:make-apparatus($node, $rdg-map)
          else
            ()
        )

    case element(tei:lb) return
      if ($node/preceding-sibling::*[1][self::tei:titlePart
        or self::tei:head]) then
        ()
      else
        <br />

    case element(tei:choice) return
      (if ($node/tei:abbr) then
        critical:transform($node/tei:abbr, true(), $rdg-map)
      else if ($node/tei:corr) then
        critical:transform($node/tei:corr, true(), $rdg-map)
      (: TODO: fully implement case with seg:)
      else if ($node/tei:supplied) then
        critical:transform($node/tei:orig, true(), $rdg-map)
      else
        (),
      commons:make-whitespace-after($node))

    case element(tei:hi) return
      if ($node/@rend) then
        (element xhtml:span
        {
          if ($node/@rend = "spaced-out"
          and matches(substring($node/following::text()[1], 1, 1), "[,!\?\.:;]")) then
              attribute class {"spaced-out spaced-out-no-right-padding"}
          else
              attribute class {$node/@rend},
          critical:transform($node/node(), true(), $rdg-map)
        },
        commons:make-whitespace-after($node))
      else
        (<em>{critical:transform($node/node(), true(), $rdg-map)}</em>,
        commons:make-whitespace-after($node))

    case element(tei:aligned) return
      element xhtml:span
      {
        attribute class {$node/@rend},
        critical:transform($node/node(), true(), $rdg-map)
      }

    case element(tei:label) return
      if ($node/@type = "headword") then
        (<b>{critical:transform($node/node(), true(), $rdg-map)}</b>,
        commons:make-whitespace-after($node))
      else
        (element xhtml:span
        {
          attribute class {"label-wrapper"},
          critical:transform($node/node(), true(), $rdg-map)
        },
        commons:make-whitespace-after($node))

    case element(tei:list) return
      element xhtml:ul
      {
        (if ($node/parent::tei:item) then
          attribute class {"nested-list"}

        else if ($node/ancestor::tei:rdg) then
          attribute class {"rdg-list"}

        else
          attribute class {"text-list"}),
          critical:transform($node/node(), true(), $rdg-map)
      }

    case element(tei:item) return
      element xhtml:li
      {
        (if ($node/ancestor::tei:div[@type = "contents"]) then
          attribute class {"list-item-toc"}
        else
          attribute class {"list-item"}),
        critical:transform($node/node(), true(), $rdg-map)
      }

    case element(tei:milestone) return
      let $edition := replace(replace($node/@edRef, "#", ""), " ", "")
      return
      if ($node/@unit = "fn-break") then
        element xhtml:span
        {
          attribute class {"fn-break"},
          string("|" || $edition || $node/@n || "|")
        }
      else
        let $edition := replace($node/@edRef, "[# ]+", "")
        let $ancestor-rdg-type := $node/ancestor::tei:rdg[1]/@type
        return
          if ($ancestor-rdg-type = ("ppl", "ptl")
          and ($node/preceding-sibling::*[1][not(self::tei:pb
            or self::tei:list
            or self::tei:rdgMarker)]
            (:this leads to some br that aren't allowed:)
(:            or $node/parent::tei:seg/preceding-sibling::node())) then:)
            )) then
            (<br />,
            if ($node/@unit = "p") then
              <span class="filler" />
            else
              ())
          else
            (element xhtml:span
            {
              attribute class {"milestone"},
              switch($node/@unit)
                case "p" return "⌇⌇"
                case "no-p" return <s>⌇⌇</s>
                case "line" return "⌇"
                case "no-line" return <s>⌇</s>
                default return "",
              <sup>{$edition}</sup>
            },
            (: tei:milestone never appears within a word :)
            text {" "})


    case element(tei:ptr) return
      commons:make-whitespace-after($node)

    case element(tei:index) return
      commons:make-whitespace-after($node)

    case element(tei:citedRange) return
      (element xhtml:span
      {
        attribute class {"biblical-refernce"},
        critical:transform($node/node(), true(), $rdg-map)
      },
      commons:make-whitespace-after($node))

    case element(tei:foreign) return
      if ($node/@xml:lang ="gr") then
        (element xhtml:span
        {
          attribute class {"hi-greek"},
          critical:transform($node/node(), true(), $rdg-map)
        },
        commons:make-whitespace-after($node))
      else
        (critical:transform($node/node(), true(), $rdg-map),
        commons:make-whitespace-after($node))

    case element(tei:line) return
      element xhtml:span
      {
        attribute class {"line"},
        critical:transform($node/node(), true(), $rdg-map)
      }

    case element(tei:lg) return
      element xhtml:div
      {
        attribute class {"linegroup"},
        critical:transform($node/node(), true(), $rdg-map)
      }

    case element(tei:row) return
      if ($node/preceding-sibling::tei:row) then
        <table><tr>
          {critical:transform($node/node(), true(), $rdg-map)}
        </tr></table>
      else
        <tr>{critical:transform($node/node(), true(), $rdg-map)}</tr>

    case element(tei:cell) return
      <td>{critical:transform($node/node(), true(), $rdg-map)}</td>

    case element(tei:app) return
      if ($node/@type = "structural-variance") then
        if ($node/tei:lem[@type = "missing-structure"]) then
          ()
        else if ($node/tei:rdg[@type = "varying-structure"]) then
          critical:transform($node/tei:lem, true(), $rdg-map)
        else
          ()
      else
        (critical:transform($node/node(), true(), $rdg-map),
        commons:make-whitespace-after($node))

    case element(tei:rdgMarker) return
      if ($display-markers) then
        if ($node/@type = ("ppl", "ptl")
        and $node/@context = "rdg") then
          (critical:make-marker($node/@wit, false(), $node/@mark),
          if ($node/following-sibling::*[1][self::tei:rdgMarker]
          and ($node/following-sibling::node()[1][self::tei:rdgMarker]
            or normalize-space($node/following-sibling::node()[1]) = "")) then
            text{" "}
          else
            ())

        else if ($node/@type = ("ppl", "pp")
        and $node/@mark = ("open", "close")
        and $node/@context = "lem") then
          critical:make-marker($node/@wit, true(), $node/@mark)

        else if ($node/@type = "om") then
          critical:make-marker($node/@wit, true(), $node/@mark)

        else if ($node/@type = "v"
        and $node/@context = "lem"
        and $node/@mark = "close") then
          for $ref in tokenize($node/@ref, " ")
            let $corresp := $node/ancestor::*[last()]//tei:rdg[@id = $ref]
            return
              critical:make-variant($corresp, $rdg-map)

        else
          ()
      else
        ()

    case element(tei:rdg) return
      if ($node[@type = ("ppl", "ptl")]) then
        element xhtml:div
        {
          attribute class {"rdg-crit"},
          critical:transform($node/node(), true(), $rdg-map)
        }
      else if ($node[@type = ("pt", "pp")]
      and not($node/preceding-sibling::tei:rdg/@type = $node/@type)) then
        critical:make-link($node, $rdg-map)
      else
        ()

    case element(tei:pb) return
      (commons:make-whitespace-before($node),
      element xhtml:span
      {
        attribute{"class"}{"pb"},
        concat(
          string("|" || replace($node/@edRef, "[# ]+", "")),
          (if ($node/@type = "sp") then string("[") else ""),
          string($node/@n),
          (if ($node/@type = "sp") then string("]") else ""),
          string("|")
        )
        },
        commons:make-whitespace-after($node))

    case element(tei:cb) return
      element xhtml:span
      {
        attribute{"class"}{"cb"},
        concat(
          string("|" || replace($node/@edRef, "[# ]+", "")),
          string($node/@n),
          string("|")
        ),
        commons:make-whitespace-after($node)
      }

    case element(tei:supplied) return
      (text {"["},
      critical:transform($node/node(), $display-markers, $rdg-map),
      text {"]"},
      commons:make-whitespace-after($node))

    case element(tei:seg) return
        if ($node[@type = "margin"]) then
            element xhtml:span
            {
                attribute class {"margin"},
                text{"{"},
                critical:transform($node/node(), true(), $rdg-map),
                text{"}"}
            }
        else
            (critical:transform($node/node(), true(), $rdg-map),
            commons:make-whitespace-after($node))

    case element (tei:floatingText) return
        element xhtml:div
        {
            attribute class {"floating-text"},
            critical:transform($node/node(), true(), $rdg-map)
        }

    case element (tei:signed) return
        element xhtml:div
        {
            attribute class {"signed"},
            critical:transform($node/node(), true(), $rdg-map)
        }

    default return
      (critical:transform($node/node(), true(), $rdg-map),
      commons:make-whitespace-after($node))
};

(:~
 : Creates a xhtml:span for a tei:rdg which is relevant for textual criticism
 : and which is not handled in the main text, i.e. tei:rdg[@type = ("v", "pp", "pt")].
 : All xhtml:spans of a tei:div[@type = "section"] are displayed en bloc at the
 : end of the respective div.
 :
 : This function is called by critical:make-apparatus, critical:make-apparatus-entry and critical:make-link.
 :
 : @author Michelle Weidling
 : @author Stefan Hynek
 : @version 1.0
 : @param $node the current tei:rdg
 : @return element that is the ancestor of $node
 :)

declare function critical:get-rdg-ancestor($node as element(tei:rdg)) as element(*)
{
    if ($node/ancestor::tei:app[@type = "standalone"][not(descendant::tei:div[@type])]) then
        $node/ancestor::tei:app[@type = "standalone"]
    else if ($node/ancestor::tei:head[not(ancestor::tei:div[@type = "section"])]) then
        $node/ancestor::tei:head
    else
        $node/ancestor::tei:div[@type][1]
};

(:~
 : Creates a xhtml:span for a tei:rdg which is relevant for textual criticism
 : and which is not handled in the main text, i.e. tei:rdg[@type = ("v", "pp", "pt")].
 : All xhtml:spans of a tei:div[@type = "section"] are displayed en bloc at the
 : end of the respective div.
 :
 : This function is called by critical:transformation.
 :
 : @author Hannes Riebl
 : @author Michelle Weidling
 : @version 1.0
 : @param $node the current tei:rdg
 : @return an xhtml:span with the tei:rdg's witness and the wit's text
 :)
declare function critical:make-apparatus-entry($node as element(tei:rdg),
$rdg-map as map(*)) as element(xhtml:span)
{
  let $rdg-ancestor := critical:get-rdg-ancestor($node)

  let $rdg-count := critical:get-rdg-pos($rdg-ancestor, $node, $rdg-map)
  let $id := util:hash($node/@id,"md5") || "-" || $rdg-count

  let $all-wits := string-join($node/../tei:rdg[@type = ("v", "pp", "pt")]/@wit, "")
  let $all-wits := replace($all-wits, "[# ]+", "")
  return
    element xhtml:span
    {
      attribute class {"app-content"},
      attribute id {$id},
      <strong>
        {$all-wits}
        {$rdg-count}:
      </strong>,

      if (count($node/../tei:rdg) = 2
      and $node/../tei:rdg[@type = ("om", "ppl", "ptl", "typo-correction",
      "varying-structure", "varying-target")]) then
        for $rdg in $node/../tei:rdg[@type = ("v", "pp", "pt")] return
          critical:transform($rdg/node(), false(), $rdg-map)

      else if (count($node/../tei:rdg) gt 1) then
        for $rdg in $node/../tei:rdg[@type = ("v", "pp", "pt")]
          let $wit := replace($rdg/@wit, "[# ]+", "")
          let $closing := if (not($rdg/following-sibling::tei:rdg[@type = ("v", "pp", "pt")])) then
              ")"
            else
              ");"
          let $ref := replace("(" || $wit || $closing, " ", "") || " "
          return
            (critical:transform($rdg/node(), false(), $rdg-map),
            text {" "},
            <span style="white-space: nowrap;">{$ref}</span>)

      else
        critical:transform($node/node(), false(), $rdg-map),
        text{" "}
    }
};

(:~
 : Triggers the creation of a link in the main text to the critical apparatus at
 : the end of a div for a given variant.
 :
 : This function is called by critical:transformation.
 :
 : @author Hannes Riebl
 : @author Michelle Weidling
 : @version 1.0
 : @param $node the current tei:rdg
 : @return an xhtml:span with the tei:rdg's witness and the wit's text
 :)
declare function critical:make-variant($node as element(tei:rdg), $rdg-map as map(*))
as node()*
{
  if (not($node/preceding-sibling::tei:rdg/@type = $node/@type)) then
    critical:make-link($node, $rdg-map)
  else
    ()
};


(:~
 : Creates a link in the main text to the critical apparatus at the end of a div
 : for a given variant or insertion.
 :
 : This function is called by critical:transformation.
 :
 : @author Hannes Riebl
 : @author Michelle Weidling
 : @version 1.0
 : @param $node the current tei:rdg
 : @return an xhtml:span with the tei:rdg's witness and the wit's text
 :)
declare function critical:make-link($node as element(tei:rdg), $rdg-map as map(*))
as element(xhtml:span)
{
  let $rdg-ancestor := critical:get-rdg-ancestor($node)

  let $rdg-count := critical:get-rdg-pos($rdg-ancestor, $node, $rdg-map)
  let $type := $node/@type
  let $wits := string-join($node/parent::tei:app/tei:rdg[@type = $type]/@wit, "")
  let $wits := replace($wits, "[# ]+", "")
  let $id := util:hash($node/@id, "md5") || "-" || $rdg-count

  return
    element xhtml:span
    {
      attribute class {"insertion-wrapper"},
      (if ($node[@type = ("pt", "pp")]) then
        text {" ∥"}
      else
        ()),
      element xhtml:a
      {
        attribute href {"#" || $id},
        <sup>{$wits}{$rdg-count}</sup>
      },
      (if (commons:starts-next-text-with-punctuation($node, "app")) then
        ()
      else
        text {" "})
    }
};

declare function critical:make-apparatus($node as element(), $rdg-map as map(*))
as element(xhtml:span)
{
  let $apps := $node//tei:app[tei:rdg[@type = ("v", "pt", "pp")]]
  let $rdgs := $apps//tei:rdg[@type = ("v", "pt", "pp")]
  return
    element xhtml:span
    {
      attribute class {"critical-apparatus"},
      for $rdg in $rdgs
        let $rdg-ancestor := critical:get-rdg-ancestor($rdg)
        let $rdg-number := critical:get-rdg-pos($rdg-ancestor, $rdg, $rdg-map)
        order by number($rdg-number)
        return
          if (not($rdg/preceding-sibling::tei:rdg[@type = ("v", "pt", "pp")])) then
            critical:make-apparatus-entry($rdg, $rdg-map)
          else
            ()
    }
};


(:~
 : Retrieves the number/position of a tei:rdg[@type = ("v", "pt", "pp")] within
 : its ancestor::tei:div[@type].
 :
 : @author Michelle Weidling
 : @see xml-utils.xqm/xml-utils:make-rdg-map($resource as node()*) for information about the map design
 : @param $div the current tei:rdg's ancestor::tei:div[@type]
 : @param $rdg the current tei:rdg[@type = ("v", "pt", "pp")]
 : @param $rdg-map
 : @return the $rdg's number/position
 : :)
declare function critical:get-rdg-pos( $ancestor-with-id as element(*), $rdg as element(tei:rdg), $rdg-map as map(*) ) as xs:double
{
    let $ancestor-key := $ancestor-with-id/@id
    let $rdg-key := $rdg/@id
    return
        map:get(map:get($rdg-map, $ancestor-key), $rdg-key)
};

(:~
 : Creates the scribal abbreviations.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $wits The sigla for one or more printing(s)
 : @param $slash A Boolean parameter indicating whether the mark needs a slash
 : @param $mark A flag indicating whether the siglum is an opening or closing one
 : @return element(xhtml:sub) The finished siglum
 :)
declare function critical:make-marker($wits as xs:string, $slash as xs:boolean,
$mark as xs:string) as element(xhtml:sup)
{
  let $open :=
    if ($slash and $mark = "open") then
      "/"
    else
      ()
  let $close :=
    if ($slash and $mark = "close") then
      "\"
    else
      ()
  let $content := $open || $wits || $close

  return
    element xhtml:sup
    {
      (: in cases or several printings these are divided by a whitespace :)
      replace($content, " ", "")
    }
};
